@@ -99,10 +99,7 @@ static bool output_attach_render(struct wlr_output *wlr_output,
 	return wlr_egl_make_current(&x11->egl, output->surf, buffer_age);
 }
 
-static bool output_commit(struct wlr_output *wlr_output) {
-	struct wlr_x11_output *output = get_x11_output_from_output(wlr_output);
-	struct wlr_x11_backend *x11 = output->x11;
-
+static bool output_test(struct wlr_output *wlr_output) {
 	if (wlr_output->pending.committed & WLR_OUTPUT_STATE_ENABLED) {
 		wlr_log(WLR_DEBUG, "Cannot disable an X11 output");
 		return false;
@@ -110,6 +107,20 @@ static bool output_commit(struct wlr_output *wlr_output) {
 
 	if (wlr_output->pending.committed & WLR_OUTPUT_STATE_MODE) {
 		assert(wlr_output->pending.mode_type == WLR_OUTPUT_STATE_MODE_CUSTOM);
+	}
+
+	return true;
+}
+
+static bool output_commit(struct wlr_output *wlr_output) {
+	struct wlr_x11_output *output = get_x11_output_from_output(wlr_output);
+	struct wlr_x11_backend *x11 = output->x11;
+
+	if (!output_test(wlr_output)) {
+		return false;
+	}
+
+	if (wlr_output->pending.committed & WLR_OUTPUT_STATE_MODE) {
 		if (!output_set_custom_mode(wlr_output,
 				wlr_output->pending.custom_mode.width,
 				wlr_output->pending.custom_mode.height,
@@ -118,6 +129,21 @@ static bool output_commit(struct wlr_output *wlr_output) {
 		}
 	}
 
+	if (wlr_output->pending.committed & WLR_OUTPUT_STATE_ADAPTIVE_SYNC_ENABLED &&
+			x11->atoms.variable_refresh != XCB_ATOM_NONE) {
+		if (wlr_output->pending.adaptive_sync_enabled) {
+			uint32_t enabled = 1;
+			xcb_change_property(x11->xcb, XCB_PROP_MODE_REPLACE, output->win,
+				x11->atoms.variable_refresh, XCB_ATOM_CARDINAL, 32, 1,
+				&enabled);
+			wlr_output->adaptive_sync_status = WLR_OUTPUT_ADAPTIVE_SYNC_UNKNOWN;
+		} else {
+			xcb_delete_property(x11->xcb, output->win,
+				x11->atoms.variable_refresh);
+			wlr_output->adaptive_sync_status = WLR_OUTPUT_ADAPTIVE_SYNC_DISABLED;
+		}
+	}
+
 	if (wlr_output->pending.committed & WLR_OUTPUT_STATE_BUFFER) {
 		pixman_region32_t *damage = NULL;
 		if (wlr_output->pending.committed & WLR_OUTPUT_STATE_DAMAGE) {
@@ -134,10 +160,17 @@ static bool output_commit(struct wlr_output *wlr_output) {
 	return true;
 }
 
+static void output_rollback_render(struct wlr_output *wlr_output) {
+	struct wlr_x11_output *output = get_x11_output_from_output(wlr_output);
+	wlr_egl_unset_current(&output->x11->egl);
+}
+
 static const struct wlr_output_impl output_impl = {
 	.destroy = output_destroy,
 	.attach_render = output_attach_render,
+	.test = output_test,
 	.commit = output_commit,
+	.rollback_render = output_rollback_render,
 };
 
 struct wlr_output *wlr_x11_output_create(struct wlr_backend *backend) {
