@@ -25,24 +25,37 @@ struct wlr_drm_plane {
 	struct wlr_drm_surface surf;
 	struct wlr_drm_surface mgpu_surf;
 
+	/* Buffer to be submitted to the kernel on the next page-flip */
+	struct wlr_drm_fb pending_fb;
+	/* Buffer submitted to the kernel, will be presented on next vblank */
+	struct wlr_drm_fb queued_fb;
+	/* Buffer currently displayed on screen */
+	struct wlr_drm_fb current_fb;
+
 	uint32_t drm_format; // ARGB8888 or XRGB8888
 	struct wlr_drm_format_set formats;
 
 	// Only used by cursor
-	float matrix[9];
 	bool cursor_enabled;
 	int32_t cursor_hotspot_x, cursor_hotspot_y;
 
 	union wlr_drm_plane_props props;
 };
 
+struct wlr_drm_crtc_state {
+	bool active;
+	struct wlr_drm_mode *mode;
+};
+
 struct wlr_drm_crtc {
 	uint32_t id;
 
+	bool pending_modeset;
+	struct wlr_drm_crtc_state pending, current;
+
 	// Atomic modesetting only
 	uint32_t mode_id;
 	uint32_t gamma_lut;
-	drmModeAtomicReq *atomic;
 
 	// Legacy only
 	drmModeCrtc *legacy_crtc;
@@ -58,11 +71,6 @@ struct wlr_drm_crtc {
 	uint32_t *overlays;
 
 	union wlr_drm_crtc_props props;
-
-	struct wl_list connectors;
-
-	uint16_t *gamma_table;
-	size_t gamma_table_size;
 };
 
 struct wlr_drm_backend {
@@ -119,23 +127,19 @@ struct wlr_drm_connector {
 
 	union wlr_drm_connector_props props;
 
-	uint32_t width, height;
 	int32_t cursor_x, cursor_y;
 
 	drmModeCrtc *old_crtc;
 
-	bool pageflip_pending;
-	struct wl_event_source *retry_pageflip;
 	struct wl_list link;
 
-	// DMA-BUF to be displayed on next commit
-	struct wlr_dmabuf_attributes pending_dmabuf;
-	// Buffer submitted to the kernel but not yet displayed
-	struct wlr_buffer *pending_buffer;
-	struct gbm_bo *pending_bo;
-	// Buffer currently being displayed
-	struct wlr_buffer *current_buffer;
-	struct gbm_bo *current_bo;
+	/*
+	 * We've asked for a state change in the kernel, and yet to recieve a
+	 * notification for its completion. Currently, the kernel only has a
+	 * queue length of 1, and no way to modify your submissions after
+	 * they're sent.
+	 */
+	bool pageflip_pending;
 };
 
 struct wlr_drm_backend *get_drm_backend_from_backend(
@@ -146,15 +150,13 @@ void finish_drm_resources(struct wlr_drm_backend *drm);
 void restore_drm_outputs(struct wlr_drm_backend *drm);
 void scan_drm_connectors(struct wlr_drm_backend *state);
 int handle_drm_event(int fd, uint32_t mask, void *data);
-bool enable_drm_connector(struct wlr_output *output, bool enable);
-bool set_drm_connector_gamma(struct wlr_output *output, size_t size,
-	const uint16_t *r, const uint16_t *g, const uint16_t *b);
-bool drm_connector_set_mode(struct wlr_output *output,
+bool drm_connector_set_mode(struct wlr_drm_connector *conn,
 	struct wlr_output_mode *mode);
+bool drm_connector_is_cursor_visible(struct wlr_drm_connector *conn);
+bool drm_connector_supports_vrr(struct wlr_drm_connector *conn);
+size_t drm_crtc_get_gamma_lut_size(struct wlr_drm_backend *drm,
+	struct wlr_drm_crtc *crtc);
 
-bool legacy_crtc_set_cursor(struct wlr_drm_backend *drm,
-	struct wlr_drm_crtc *crtc, struct gbm_bo *bo);
-bool legacy_crtc_move_cursor(struct wlr_drm_backend *drm,
-	struct wlr_drm_crtc *crtc, int x, int y);
+struct wlr_drm_fb *plane_get_next_fb(struct wlr_drm_plane *plane);
 
 #endif
