@@ -16,6 +16,7 @@ union wlr_drm_connector_props {
 		uint32_t dpms;
 		uint32_t link_status; // not guaranteed to exist
 		uint32_t path;
+		uint32_t vrr_capable; // not guaranteed to exist
 
 		// atomic-modesetting only
 
@@ -29,13 +30,14 @@ union wlr_drm_crtc_props {
 		// Neither of these are guaranteed to exist
 		uint32_t rotation;
 		uint32_t scaling_mode;
+		uint32_t vrr_enabled;
+		uint32_t gamma_lut;
+		uint32_t gamma_lut_size;
 
 		// atomic-modesetting only
 
 		uint32_t active;
 		uint32_t mode_id;
-		uint32_t gamma_lut;
-		uint32_t gamma_lut_size;
 	};
 	uint32_t props[6];
 };
