@@ -18,6 +18,11 @@ struct wlr_box {
 	int width, height;
 };
 
+struct wlr_fbox {
+	double x, y;
+	double width, height;
+};
+
 void wlr_box_closest_point(const struct wlr_box *box, double x, double y,
 	double *dest_x, double *dest_y);
 
