@@ -10,9 +10,17 @@
 #include <wlr/render/interface.h>
 #include <wlr/render/wlr_renderer.h>
 #include <wlr/types/wlr_matrix.h>
+#include <wlr/types/wlr_linux_dmabuf_v1.h>
 #include <wlr/util/log.h>
 #include "render/gles2.h"
 
+static const GLfloat verts[] = {
+	1, 0, // top right
+	0, 0, // top left
+	1, 1, // bottom right
+	0, 1, // bottom left
+};
+
 struct wlr_gles2_procs gles2_procs = {0};
 
 static const struct wlr_renderer_impl renderer_impl;
@@ -85,34 +93,9 @@ static void gles2_scissor(struct wlr_renderer *wlr_renderer,
 	POP_GLES2_DEBUG;
 }
 
-static void draw_quad(void) {
-	GLfloat verts[] = {
-		1, 0, // top right
-		0, 0, // top left
-		1, 1, // bottom right
-		0, 1, // bottom left
-	};
-	GLfloat texcoord[] = {
-		1, 0, // top right
-		0, 0, // top left
-		1, 1, // bottom right
-		0, 1, // bottom left
-	};
-
-	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, verts);
-	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, texcoord);
-
-	glEnableVertexAttribArray(0);
-	glEnableVertexAttribArray(1);
-
-	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
-
-	glDisableVertexAttribArray(0);
-	glDisableVertexAttribArray(1);
-}
-
-static bool gles2_render_texture_with_matrix(struct wlr_renderer *wlr_renderer,
-		struct wlr_texture *wlr_texture, const float matrix[static 9],
+static bool gles2_render_subtexture_with_matrix(
+		struct wlr_renderer *wlr_renderer, struct wlr_texture *wlr_texture,
+		const struct wlr_fbox *box, const float matrix[static 9],
 		float alpha) {
 	struct wlr_gles2_renderer *renderer =
 		gles2_get_renderer_in_context(wlr_renderer);
@@ -161,7 +144,27 @@ static bool gles2_render_texture_with_matrix(struct wlr_renderer *wlr_renderer,
 	glUniform1i(shader->tex, 0);
 	glUniform1f(shader->alpha, alpha);
 
-	draw_quad();
+	const GLfloat x1 = box->x / wlr_texture->width;
+	const GLfloat y1 = box->y / wlr_texture->height;
+	const GLfloat x2 = (box->x + box->width) / wlr_texture->width;
+	const GLfloat y2 = (box->y + box->height) / wlr_texture->height;
+	const GLfloat texcoord[] = {
+		x2, y1, // top right
+		x1, y1, // top left
+		x2, y2, // bottom right
+		x1, y2, // bottom left
+	};
+
+	glVertexAttribPointer(shader->pos_attrib, 2, GL_FLOAT, GL_FALSE, 0, verts);
+	glVertexAttribPointer(shader->tex_attrib, 2, GL_FLOAT, GL_FALSE, 0, texcoord);
+
+	glEnableVertexAttribArray(shader->pos_attrib);
+	glEnableVertexAttribArray(shader->tex_attrib);
+
+	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+	glDisableVertexAttribArray(shader->pos_attrib);
+	glDisableVertexAttribArray(shader->tex_attrib);
 
 	glBindTexture(texture->target, 0);
 
@@ -169,7 +172,6 @@ static bool gles2_render_texture_with_matrix(struct wlr_renderer *wlr_renderer,
 	return true;
 }
 
-
 static void gles2_render_quad_with_matrix(struct wlr_renderer *wlr_renderer,
 		const float color[static 4], const float matrix[static 9]) {
 	struct wlr_gles2_renderer *renderer =
@@ -185,7 +187,16 @@ static void gles2_render_quad_with_matrix(struct wlr_renderer *wlr_renderer,
 
 	glUniformMatrix3fv(renderer->shaders.quad.proj, 1, GL_FALSE, transposition);
 	glUniform4f(renderer->shaders.quad.color, color[0], color[1], color[2], color[3]);
-	draw_quad();
+
+	glVertexAttribPointer(renderer->shaders.quad.pos_attrib, 2, GL_FLOAT, GL_FALSE,
+			0, verts);
+
+	glEnableVertexAttribArray(renderer->shaders.quad.pos_attrib);
+
+	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+	glDisableVertexAttribArray(renderer->shaders.quad.pos_attrib);
+
 	POP_GLES2_DEBUG;
 }
 
@@ -199,12 +210,31 @@ static void gles2_render_ellipse_with_matrix(struct wlr_renderer *wlr_renderer,
 	float transposition[9];
 	wlr_matrix_transpose(transposition, matrix);
 
+	static const GLfloat texcoord[] = {
+		1, 0, // top right
+		0, 0, // top left
+		1, 1, // bottom right
+		0, 1, // bottom left
+	};
+
 	PUSH_GLES2_DEBUG;
 	glUseProgram(renderer->shaders.ellipse.program);
 
 	glUniformMatrix3fv(renderer->shaders.ellipse.proj, 1, GL_FALSE, transposition);
 	glUniform4f(renderer->shaders.ellipse.color, color[0], color[1], color[2], color[3]);
-	draw_quad();
+
+	glVertexAttribPointer(renderer->shaders.ellipse.pos_attrib, 2, GL_FLOAT,
+			GL_FALSE, 0, verts);
+	glVertexAttribPointer(renderer->shaders.ellipse.tex_attrib, 2, GL_FLOAT,
+			GL_FALSE, 0, texcoord);
+
+	glEnableVertexAttribArray(renderer->shaders.ellipse.pos_attrib);
+	glEnableVertexAttribArray(renderer->shaders.ellipse.tex_attrib);
+
+	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
+
+	glDisableVertexAttribArray(renderer->shaders.ellipse.pos_attrib);
+	glDisableVertexAttribArray(renderer->shaders.ellipse.tex_attrib);
 	POP_GLES2_DEBUG;
 }
 
@@ -330,6 +360,88 @@ static bool gles2_read_pixels(struct wlr_renderer *wlr_renderer,
 	return glGetError() == GL_NO_ERROR;
 }
 
+static bool gles2_blit_dmabuf(struct wlr_renderer *wlr_renderer,
+		struct wlr_dmabuf_attributes *dst_attr,
+		struct wlr_dmabuf_attributes *src_attr) {
+	if (!gles2_procs.glEGLImageTargetRenderbufferStorageOES) {
+		return false;
+	}
+
+	struct wlr_egl_context old_context;
+	wlr_egl_save_context(&old_context);
+
+	bool r = false;
+	struct wlr_texture *src_tex =
+		wlr_texture_from_dmabuf(wlr_renderer, src_attr);
+	if (!src_tex) {
+		goto restore_context_out;
+	}
+
+	// This is to take into account y-inversion on both buffers rather than
+	// just the source buffer.
+	bool src_inverted_y =
+		!!(src_attr->flags & WLR_DMABUF_ATTRIBUTES_FLAGS_Y_INVERT);
+	bool dst_inverted_y =
+		!!(dst_attr->flags & WLR_DMABUF_ATTRIBUTES_FLAGS_Y_INVERT);
+	struct wlr_gles2_texture *gles2_src_tex = gles2_get_texture(src_tex);
+	// The result is negated because wlr_matrix_projection y-inverts the
+	// texture.
+	gles2_src_tex->inverted_y = !(src_inverted_y ^ dst_inverted_y);
+
+	struct wlr_egl *egl = wlr_gles2_renderer_get_egl(wlr_renderer);
+	if (!wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL)) {
+		goto texture_destroy_out;
+	}
+
+	// TODO: The imported buffer should be checked with
+	// eglQueryDmaBufModifiersEXT to see if it may be modified.
+	bool external_only = false;
+	EGLImageKHR image = wlr_egl_create_image_from_dmabuf(egl, dst_attr,
+			&external_only);
+	if (image == EGL_NO_IMAGE_KHR) {
+		goto texture_destroy_out;
+	}
+
+	GLuint rbo = 0;
+	glGenRenderbuffers(1, &rbo);
+	glBindRenderbuffer(GL_RENDERBUFFER, rbo);
+	gles2_procs.glEGLImageTargetRenderbufferStorageOES(GL_RENDERBUFFER,
+			image);
+	glBindRenderbuffer(GL_RENDERBUFFER, 0);
+
+	GLuint fbo = 0;
+	glGenFramebuffers(1, &fbo);
+	glBindFramebuffer(GL_FRAMEBUFFER, fbo);
+	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
+			GL_RENDERBUFFER, rbo);
+
+	GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
+	if (status != GL_FRAMEBUFFER_COMPLETE) {
+		goto out;
+	}
+
+	// TODO: use ANGLE_framebuffer_blit if available
+	float mat[9];
+	wlr_matrix_projection(mat, 1, 1, WL_OUTPUT_TRANSFORM_NORMAL);
+
+	wlr_renderer_begin(wlr_renderer, dst_attr->width, dst_attr->height);
+	wlr_renderer_clear(wlr_renderer, (float[]){ 0.0, 0.0, 0.0, 0.0 });
+	wlr_render_texture_with_matrix(wlr_renderer, src_tex, mat, 1.0f);
+	wlr_renderer_end(wlr_renderer);
+
+	r = true;
+out:
+	glBindFramebuffer(GL_FRAMEBUFFER, 0);
+	glDeleteFramebuffers(1, &fbo);
+	glDeleteRenderbuffers(1, &rbo);
+	wlr_egl_destroy_image(egl, image);
+texture_destroy_out:
+	wlr_texture_destroy(src_tex);
+restore_context_out:
+	wlr_egl_restore_context(&old_context);
+	return r;
+}
+
 static struct wlr_texture *gles2_texture_from_pixels(
 		struct wlr_renderer *wlr_renderer, enum wl_shm_format wl_fmt,
 		uint32_t stride, uint32_t width, uint32_t height, const void *data) {
@@ -351,13 +463,29 @@ static struct wlr_texture *gles2_texture_from_dmabuf(
 	return wlr_gles2_texture_from_dmabuf(renderer->egl, attribs);
 }
 
-static void gles2_init_wl_display(struct wlr_renderer *wlr_renderer,
+static bool gles2_init_wl_display(struct wlr_renderer *wlr_renderer,
 		struct wl_display *wl_display) {
 	struct wlr_gles2_renderer *renderer =
 		gles2_get_renderer(wlr_renderer);
-	if (!wlr_egl_bind_display(renderer->egl, wl_display)) {
-		wlr_log(WLR_INFO, "failed to bind wl_display to EGL");
+
+	if (renderer->egl->exts.bind_wayland_display_wl) {
+		if (!wlr_egl_bind_display(renderer->egl, wl_display)) {
+			wlr_log(WLR_ERROR, "Failed to bind wl_display to EGL");
+			return false;
+		}
+	} else {
+		wlr_log(WLR_INFO, "EGL_WL_bind_wayland_display is not supported");
 	}
+
+	if (renderer->egl->exts.image_dmabuf_import_ext) {
+		if (wlr_linux_dmabuf_v1_create(wl_display, wlr_renderer) == NULL) {
+			return false;
+		}
+	} else {
+		wlr_log(WLR_INFO, "EGL_EXT_image_dma_buf_import is not supported");
+	}
+
+	return true;
 }
 
 struct wlr_egl *wlr_gles2_renderer_get_egl(struct wlr_renderer *wlr_renderer) {
@@ -384,6 +512,8 @@ static void gles2_destroy(struct wlr_renderer *wlr_renderer) {
 		gles2_procs.glDebugMessageCallbackKHR(NULL, NULL);
 	}
 
+	wlr_egl_unset_current(renderer->egl);
+
 	free(renderer);
 }
 
@@ -393,7 +523,7 @@ static const struct wlr_renderer_impl renderer_impl = {
 	.end = gles2_end,
 	.clear = gles2_clear,
 	.scissor = gles2_scissor,
-	.render_texture_with_matrix = gles2_render_texture_with_matrix,
+	.render_subtexture_with_matrix = gles2_render_subtexture_with_matrix,
 	.render_quad_with_matrix = gles2_render_quad_with_matrix,
 	.render_ellipse_with_matrix = gles2_render_ellipse_with_matrix,
 	.formats = gles2_renderer_formats,
@@ -407,6 +537,7 @@ static const struct wlr_renderer_impl renderer_impl = {
 	.texture_from_wl_drm = gles2_texture_from_wl_drm,
 	.texture_from_dmabuf = gles2_texture_from_dmabuf,
 	.init_wl_display = gles2_init_wl_display,
+	.blit_dmabuf = gles2_blit_dmabuf,
 };
 
 void push_gles2_marker(const char *file, const char *func) {
@@ -539,6 +670,10 @@ extern const GLchar tex_fragment_src_rgbx[];
 extern const GLchar tex_fragment_src_external[];
 
 struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
+	if (!wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL)) {
+		return NULL;
+	}
+
 	const char *exts_str = (const char *)glGetString(GL_EXTENSIONS);
 	if (exts_str == NULL) {
 		wlr_log(WLR_ERROR, "Failed to get GL_EXTENSIONS");
@@ -553,10 +688,7 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	wlr_renderer_init(&renderer->wlr_renderer, &renderer_impl);
 
 	renderer->egl = egl;
-	if (!wlr_egl_make_current(renderer->egl, EGL_NO_SURFACE, NULL)) {
-		free(renderer);
-		return NULL;
-	}
+	renderer->exts_str = exts_str;
 
 	wlr_log(WLR_INFO, "Using %s", glGetString(GL_VERSION));
 	wlr_log(WLR_INFO, "GL vendor: %s", glGetString(GL_VENDOR));
@@ -586,6 +718,12 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 			"glEGLImageTargetTexture2DOES");
 	}
 
+	if (check_gl_ext(exts_str, "GL_OES_EGL_image")) {
+		renderer->exts.egl_image_oes = true;
+		load_gl_proc(&gles2_procs.glEGLImageTargetRenderbufferStorageOES,
+			"glEGLImageTargetRenderbufferStorageOES");
+	}
+
 	if (renderer->exts.debug_khr) {
 		glEnable(GL_DEBUG_OUTPUT_KHR);
 		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS_KHR);
@@ -608,6 +746,7 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	}
 	renderer->shaders.quad.proj = glGetUniformLocation(prog, "proj");
 	renderer->shaders.quad.color = glGetUniformLocation(prog, "color");
+	renderer->shaders.quad.pos_attrib = glGetAttribLocation(prog, "pos");
 
 	renderer->shaders.ellipse.program = prog =
 		link_program(quad_vertex_src, ellipse_fragment_src);
@@ -616,6 +755,8 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	}
 	renderer->shaders.ellipse.proj = glGetUniformLocation(prog, "proj");
 	renderer->shaders.ellipse.color = glGetUniformLocation(prog, "color");
+	renderer->shaders.ellipse.pos_attrib = glGetAttribLocation(prog, "pos");
+	renderer->shaders.ellipse.tex_attrib = glGetAttribLocation(prog, "texcoord");
 
 	renderer->shaders.tex_rgba.program = prog =
 		link_program(tex_vertex_src, tex_fragment_src_rgba);
@@ -626,6 +767,8 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	renderer->shaders.tex_rgba.invert_y = glGetUniformLocation(prog, "invert_y");
 	renderer->shaders.tex_rgba.tex = glGetUniformLocation(prog, "tex");
 	renderer->shaders.tex_rgba.alpha = glGetUniformLocation(prog, "alpha");
+	renderer->shaders.tex_rgba.pos_attrib = glGetAttribLocation(prog, "pos");
+	renderer->shaders.tex_rgba.tex_attrib = glGetAttribLocation(prog, "texcoord");
 
 	renderer->shaders.tex_rgbx.program = prog =
 		link_program(tex_vertex_src, tex_fragment_src_rgbx);
@@ -636,6 +779,8 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 	renderer->shaders.tex_rgbx.invert_y = glGetUniformLocation(prog, "invert_y");
 	renderer->shaders.tex_rgbx.tex = glGetUniformLocation(prog, "tex");
 	renderer->shaders.tex_rgbx.alpha = glGetUniformLocation(prog, "alpha");
+	renderer->shaders.tex_rgbx.pos_attrib = glGetAttribLocation(prog, "pos");
+	renderer->shaders.tex_rgbx.tex_attrib = glGetAttribLocation(prog, "texcoord");
 
 	if (renderer->exts.egl_image_external_oes) {
 		renderer->shaders.tex_ext.program = prog =
@@ -647,10 +792,14 @@ struct wlr_renderer *wlr_gles2_renderer_create(struct wlr_egl *egl) {
 		renderer->shaders.tex_ext.invert_y = glGetUniformLocation(prog, "invert_y");
 		renderer->shaders.tex_ext.tex = glGetUniformLocation(prog, "tex");
 		renderer->shaders.tex_ext.alpha = glGetUniformLocation(prog, "alpha");
+		renderer->shaders.tex_ext.pos_attrib = glGetAttribLocation(prog, "pos");
+		renderer->shaders.tex_ext.tex_attrib = glGetAttribLocation(prog, "texcoord");
 	}
 
 	POP_GLES2_DEBUG;
 
+	wlr_egl_unset_current(renderer->egl);
+
 	return &renderer->wlr_renderer;
 
 error:
@@ -667,6 +816,14 @@ error:
 		gles2_procs.glDebugMessageCallbackKHR(NULL, NULL);
 	}
 
+	wlr_egl_unset_current(renderer->egl);
+
 	free(renderer);
 	return NULL;
 }
+
+bool wlr_gles2_renderer_check_ext(struct wlr_renderer *wlr_renderer,
+		const char *ext) {
+	struct wlr_gles2_renderer *renderer = gles2_get_renderer(wlr_renderer);
+	return check_gl_ext(renderer->exts_str, ext);
+}
