@@ -29,19 +29,10 @@ struct wlr_gles2_texture *gles2_get_texture(
 static struct wlr_gles2_texture *get_gles2_texture_in_context(
 		struct wlr_texture *wlr_texture) {
 	struct wlr_gles2_texture *texture = gles2_get_texture(wlr_texture);
-	if (!wlr_egl_is_current(texture->egl)) {
-		wlr_egl_make_current(texture->egl, EGL_NO_SURFACE, NULL);
-	}
+	wlr_egl_make_current(texture->egl, EGL_NO_SURFACE, NULL);
 	return texture;
 }
 
-static void gles2_texture_get_size(struct wlr_texture *wlr_texture, int *width,
-		int *height) {
-	struct wlr_gles2_texture *texture = gles2_get_texture(wlr_texture);
-	*width = texture->width;
-	*height = texture->height;
-}
-
 static bool gles2_texture_is_opaque(struct wlr_texture *wlr_texture) {
 	struct wlr_gles2_texture *texture = gles2_get_texture(wlr_texture);
 	return !texture->has_alpha;
@@ -56,6 +47,7 @@ static bool gles2_texture_write_pixels(struct wlr_texture *wlr_texture,
 
 	if (texture->target != GL_TEXTURE_2D) {
 		wlr_log(WLR_ERROR, "Cannot write pixels to immutable texture");
+		wlr_egl_unset_current(texture->egl);
 		return false;
 	}
 
@@ -82,6 +74,8 @@ static bool gles2_texture_write_pixels(struct wlr_texture *wlr_texture,
 	glBindTexture(GL_TEXTURE_2D, 0);
 
 	POP_GLES2_DEBUG;
+
+	wlr_egl_unset_current(texture->egl);
 	return true;
 }
 
@@ -110,7 +104,7 @@ static bool gles2_texture_to_dmabuf(struct wlr_texture *wlr_texture,
 	}
 
 	return wlr_egl_export_image_to_dmabuf(texture->egl, texture->image,
-		texture->width, texture->height, flags, attribs);
+		wlr_texture->width, wlr_texture->height, flags, attribs);
 }
 
 static void gles2_texture_destroy(struct wlr_texture *wlr_texture) {
@@ -118,11 +112,8 @@ static void gles2_texture_destroy(struct wlr_texture *wlr_texture) {
 		return;
 	}
 
-	struct wlr_gles2_texture *texture = gles2_get_texture(wlr_texture);
-
-	if (!wlr_egl_is_current(texture->egl)) {
-		wlr_egl_make_current(texture->egl, EGL_NO_SURFACE, NULL);
-	}
+	struct wlr_gles2_texture *texture =
+		get_gles2_texture_in_context(wlr_texture);
 
 	PUSH_GLES2_DEBUG;
 
@@ -131,11 +122,12 @@ static void gles2_texture_destroy(struct wlr_texture *wlr_texture) {
 
 	POP_GLES2_DEBUG;
 
+	wlr_egl_unset_current(texture->egl);
+
 	free(texture);
 }
 
 static const struct wlr_texture_impl texture_impl = {
-	.get_size = gles2_texture_get_size,
 	.is_opaque = gles2_texture_is_opaque,
 	.write_pixels = gles2_texture_write_pixels,
 	.to_dmabuf = gles2_texture_to_dmabuf,
@@ -145,9 +137,7 @@ static const struct wlr_texture_impl texture_impl = {
 struct wlr_texture *wlr_gles2_texture_from_pixels(struct wlr_egl *egl,
 		enum wl_shm_format wl_fmt, uint32_t stride, uint32_t width,
 		uint32_t height, const void *data) {
-	if (!wlr_egl_is_current(egl)) {
-		wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
-	}
+	wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
 
 	const struct wlr_gles2_pixel_format *fmt = get_gles2_format_from_wl(wl_fmt);
 	if (fmt == NULL) {
@@ -161,10 +151,8 @@ struct wlr_texture *wlr_gles2_texture_from_pixels(struct wlr_egl *egl,
 		wlr_log(WLR_ERROR, "Allocation failed");
 		return NULL;
 	}
-	wlr_texture_init(&texture->wlr_texture, &texture_impl);
+	wlr_texture_init(&texture->wlr_texture, &texture_impl, width, height);
 	texture->egl = egl;
-	texture->width = width;
-	texture->height = height;
 	texture->target = GL_TEXTURE_2D;
 	texture->has_alpha = fmt->has_alpha;
 	texture->wl_format = fmt->wl_format;
@@ -182,36 +170,42 @@ struct wlr_texture *wlr_gles2_texture_from_pixels(struct wlr_egl *egl,
 	glBindTexture(GL_TEXTURE_2D, 0);
 
 	POP_GLES2_DEBUG;
+
+	wlr_egl_unset_current(egl);
 	return &texture->wlr_texture;
 }
 
 struct wlr_texture *wlr_gles2_texture_from_wl_drm(struct wlr_egl *egl,
-		struct wl_resource *data) {
-	if (!wlr_egl_is_current(egl)) {
-		wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
-	}
+		struct wl_resource *resource) {
+	wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
 
 	if (!gles2_procs.glEGLImageTargetTexture2DOES) {
 		return NULL;
 	}
 
+	EGLint fmt;
+	int width, height;
+	bool inverted_y;
+	EGLImageKHR image = wlr_egl_create_image_from_wl_drm(egl, resource,
+		&fmt, &width, &height, &inverted_y);
+	if (image == EGL_NO_IMAGE_KHR) {
+		wlr_log(WLR_ERROR, "Failed to create EGL image from wl_drm resource");
+		return NULL;
+	}
+
 	struct wlr_gles2_texture *texture =
 		calloc(1, sizeof(struct wlr_gles2_texture));
 	if (texture == NULL) {
 		wlr_log(WLR_ERROR, "Allocation failed");
+		wlr_egl_destroy_image(egl, image);
 		return NULL;
 	}
-	wlr_texture_init(&texture->wlr_texture, &texture_impl);
+	wlr_texture_init(&texture->wlr_texture, &texture_impl, width, height);
 	texture->egl = egl;
 
-	EGLint fmt;
 	texture->wl_format = 0xFFFFFFFF; // texture can't be written anyways
-	texture->image = wlr_egl_create_image_from_wl_drm(egl, data, &fmt,
-		&texture->width, &texture->height, &texture->inverted_y);
-	if (texture->image == NULL) {
-		free(texture);
-		return NULL;
-	}
+	texture->image = image;
+	texture->inverted_y = inverted_y;
 
 	switch (fmt) {
 	case EGL_TEXTURE_RGB:
@@ -223,6 +217,7 @@ struct wlr_texture *wlr_gles2_texture_from_wl_drm(struct wlr_egl *egl,
 		break;
 	default:
 		wlr_log(WLR_ERROR, "Invalid or unsupported EGL buffer format");
+		wlr_egl_destroy_image(egl, image);
 		free(texture);
 		return NULL;
 	}
@@ -238,14 +233,14 @@ struct wlr_texture *wlr_gles2_texture_from_wl_drm(struct wlr_egl *egl,
 	glBindTexture(GL_TEXTURE_EXTERNAL_OES, 0);
 
 	POP_GLES2_DEBUG;
+
+	wlr_egl_unset_current(egl);
 	return &texture->wlr_texture;
 }
 
 struct wlr_texture *wlr_gles2_texture_from_dmabuf(struct wlr_egl *egl,
 		struct wlr_dmabuf_attributes *attribs) {
-	if (!wlr_egl_is_current(egl)) {
-		wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
-	}
+	wlr_egl_make_current(egl, EGL_NO_SURFACE, NULL);
 
 	if (!gles2_procs.glEGLImageTargetTexture2DOES) {
 		return NULL;
@@ -275,31 +270,35 @@ struct wlr_texture *wlr_gles2_texture_from_dmabuf(struct wlr_egl *egl,
 		wlr_log(WLR_ERROR, "Allocation failed");
 		return NULL;
 	}
-	wlr_texture_init(&texture->wlr_texture, &texture_impl);
+	wlr_texture_init(&texture->wlr_texture, &texture_impl,
+		attribs->width, attribs->height);
 	texture->egl = egl;
-	texture->width = attribs->width;
-	texture->height = attribs->height;
-	texture->target = GL_TEXTURE_EXTERNAL_OES;
 	texture->has_alpha = true;
 	texture->wl_format = 0xFFFFFFFF; // texture can't be written anyways
 	texture->inverted_y =
 		(attribs->flags & WLR_DMABUF_ATTRIBUTES_FLAGS_Y_INVERT) != 0;
 
-	texture->image = wlr_egl_create_image_from_dmabuf(egl, attribs);
-	if (texture->image == NULL) {
+	bool external_only;
+	texture->image =
+		wlr_egl_create_image_from_dmabuf(egl, attribs, &external_only);
+	if (texture->image == EGL_NO_IMAGE_KHR) {
+		wlr_log(WLR_ERROR, "Failed to create EGL image from DMA-BUF");
 		free(texture);
 		return NULL;
 	}
 
+	texture->target = external_only ? GL_TEXTURE_EXTERNAL_OES : GL_TEXTURE_2D;
+
 	PUSH_GLES2_DEBUG;
 
 	glGenTextures(1, &texture->tex);
-	glBindTexture(GL_TEXTURE_EXTERNAL_OES, texture->tex);
-	gles2_procs.glEGLImageTargetTexture2DOES(GL_TEXTURE_EXTERNAL_OES,
-		texture->image);
-	glBindTexture(GL_TEXTURE_EXTERNAL_OES, 0);
+	glBindTexture(texture->target, texture->tex);
+	gles2_procs.glEGLImageTargetTexture2DOES(texture->target, texture->image);
+	glBindTexture(texture->target, 0);
 
 	POP_GLES2_DEBUG;
+
+	wlr_egl_unset_current(egl);
 	return &texture->wlr_texture;
 }
 
