@@ -11,6 +11,7 @@
 #include <wlr/util/log.h>
 #include <wlr/util/region.h>
 #include "util/signal.h"
+#include "util/time.h"
 
 #define CALLBACK_VERSION 1
 #define SURFACE_VERSION 4
@@ -138,26 +139,57 @@ static void surface_set_input_region(struct wl_client *client,
 	}
 }
 
+/**
+ * Computes the surface viewport source size, ie. the size after applying the
+ * surface's scale, transform and cropping (via the viewport's source
+ * rectangle) but before applying the viewport scaling (via the viewport's
+ * destination rectangle).
+ */
+static void surface_state_viewport_src_size(struct wlr_surface_state *state,
+		int *out_width, int *out_height) {
+	if (state->buffer_width == 0 && state->buffer_height == 0) {
+		*out_width = *out_height = 0;
+		return;
+	}
+
+	if (state->viewport.has_src) {
+		*out_width = state->viewport.src.width;
+		*out_height = state->viewport.src.height;
+	} else {
+		int width = state->buffer_width / state->scale;
+		int height = state->buffer_height / state->scale;
+		if ((state->transform & WL_OUTPUT_TRANSFORM_90) != 0) {
+			int tmp = width;
+			width = height;
+			height = tmp;
+		}
+		*out_width = width;
+		*out_height = height;
+	}
+}
+
 static void surface_state_finalize(struct wlr_surface *surface,
 		struct wlr_surface_state *state) {
 	if ((state->committed & WLR_SURFACE_STATE_BUFFER)) {
 		if (state->buffer_resource != NULL) {
-			wlr_buffer_get_resource_size(state->buffer_resource,
+			wlr_resource_get_buffer_size(state->buffer_resource,
 				surface->renderer, &state->buffer_width, &state->buffer_height);
 		} else {
 			state->buffer_width = state->buffer_height = 0;
 		}
 	}
 
-	int width = state->buffer_width / state->scale;
-	int height = state->buffer_height / state->scale;
-	if ((state->transform & WL_OUTPUT_TRANSFORM_90) != 0) {
-		int tmp = width;
-		width = height;
-		height = tmp;
+	if (state->viewport.has_dst) {
+		if (state->buffer_width == 0 && state->buffer_height == 0) {
+			state->width = state->height = 0;
+		} else {
+			state->width = state->viewport.dst_width;
+			state->height = state->viewport.dst_height;
+		}
+	} else {
+		surface_state_viewport_src_size(state,
+			&state->width, &state->height);
 	}
-	state->width = width;
-	state->height = height;
 
 	pixman_region32_intersect_rect(&state->surface_damage,
 		&state->surface_damage, 0, 0, state->width, state->height);
@@ -182,6 +214,22 @@ static void surface_update_damage(pixman_region32_t *buffer_damage,
 		pixman_region32_init(&surface_damage);
 
 		pixman_region32_copy(&surface_damage, &pending->surface_damage);
+
+		if (pending->viewport.has_dst) {
+			int src_width, src_height;
+			surface_state_viewport_src_size(pending, &src_width, &src_height);
+			float scale_x = (float)pending->viewport.dst_width / src_width;
+			float scale_y = (float)pending->viewport.dst_height / src_height;
+			wlr_region_scale_xy(&surface_damage, &surface_damage,
+				1.0 / scale_x, 1.0 / scale_y);
+		}
+		if (pending->viewport.has_src) {
+			// This is lossy: do a best-effort conversion
+			pixman_region32_translate(&surface_damage,
+				floor(pending->viewport.src.x),
+				floor(pending->viewport.src.y));
+		}
+
 		wlr_region_transform(&surface_damage, &surface_damage,
 			wlr_output_transform_invert(pending->transform),
 			pending->width, pending->height);
@@ -229,6 +277,9 @@ static void surface_state_copy(struct wlr_surface_state *state,
 	if (next->committed & WLR_SURFACE_STATE_INPUT_REGION) {
 		pixman_region32_copy(&state->input, &next->input);
 	}
+	if (next->committed & WLR_SURFACE_STATE_VIEWPORT) {
+		memcpy(&state->viewport, &next->viewport, sizeof(state->viewport));
+	}
 
 	state->committed |= next->committed;
 }
@@ -282,29 +333,33 @@ static void surface_apply_damage(struct wlr_surface *surface) {
 	struct wl_resource *resource = surface->current.buffer_resource;
 	if (resource == NULL) {
 		// NULL commit
-		wlr_buffer_unref(surface->buffer);
+		if (surface->buffer != NULL) {
+			wlr_buffer_unlock(&surface->buffer->base);
+		}
 		surface->buffer = NULL;
 		return;
 	}
 
-	if (surface->buffer != NULL && surface->buffer->released) {
-		struct wlr_buffer *updated_buffer = wlr_buffer_apply_damage(
-			surface->buffer, resource, &surface->buffer_damage);
+	if (surface->buffer != NULL && surface->buffer->resource_released) {
+		struct wlr_client_buffer *updated_buffer =
+			wlr_client_buffer_apply_damage(surface->buffer, resource,
+			&surface->buffer_damage);
 		if (updated_buffer != NULL) {
 			surface->buffer = updated_buffer;
 			return;
 		}
 	}
 
-	wlr_buffer_unref(surface->buffer);
-	surface->buffer = NULL;
-
-	struct wlr_buffer *buffer = wlr_buffer_create(surface->renderer, resource);
+	struct wlr_client_buffer *buffer =
+		wlr_client_buffer_import(surface->renderer, resource);
 	if (buffer == NULL) {
 		wlr_log(WLR_ERROR, "Failed to upload buffer");
 		return;
 	}
 
+	if (surface->buffer != NULL) {
+		wlr_buffer_unlock(&surface->buffer->base);
+	}
 	surface->buffer = buffer;
 }
 
@@ -575,7 +630,9 @@ static void surface_handle_resource_destroy(struct wl_resource *resource) {
 	pixman_region32_fini(&surface->buffer_damage);
 	pixman_region32_fini(&surface->opaque_region);
 	pixman_region32_fini(&surface->input_region);
-	wlr_buffer_unref(surface->buffer);
+	if (surface->buffer != NULL) {
+		wlr_buffer_unlock(&surface->buffer->base);
+	}
 	free(surface);
 }
 
@@ -661,8 +718,14 @@ bool wlr_surface_set_role(struct wlr_surface *surface,
 		}
 		return false;
 	}
+	if (surface->role_data != NULL && surface->role_data != role_data) {
+		wl_resource_post_error(error_resource, error_code,
+			"Cannot reassign role %s to wl_surface@%d,"
+			"role object still exists", role->name,
+			wl_resource_get_id(surface->resource));
+		return false;
+	}
 
-	assert(surface->role_data == NULL);
 	surface->role = role;
 	surface->role_data = role_data;
 	return true;
@@ -1016,8 +1079,12 @@ struct wlr_surface *wlr_surface_surface_at(struct wlr_surface *surface,
 	}
 
 	if (wlr_surface_point_accepts_input(surface, sx, sy)) {
-		*sub_x = sx;
-		*sub_y = sy;
+		if (sub_x) {
+			*sub_x = sx;
+		}
+		if (sub_y) {
+			*sub_y = sy;
+		}
 		return surface;
 	}
 
@@ -1046,10 +1113,6 @@ void wlr_surface_send_leave(struct wlr_surface *surface,
 	}
 }
 
-static inline int64_t timespec_to_msec(const struct timespec *a) {
-	return (int64_t)a->tv_sec * 1000 + a->tv_nsec / 1000000;
-}
-
 void wlr_surface_send_frame_done(struct wlr_surface *surface,
 		const struct timespec *when) {
 	struct wl_resource *resource, *tmp;
@@ -1112,6 +1175,13 @@ void wlr_surface_get_extends(struct wlr_surface *surface, struct wlr_box *box) {
 	box->height = acc.max_y - acc.min_y;
 }
 
+static void crop_region(pixman_region32_t *dst, pixman_region32_t *src,
+		const struct wlr_box *box) {
+	pixman_region32_intersect_rect(dst, src,
+		box->x, box->y, box->width, box->height);
+	pixman_region32_translate(dst, -box->x, -box->y);
+}
+
 void wlr_surface_get_effective_damage(struct wlr_surface *surface,
 		pixman_region32_t *damage) {
 	pixman_region32_clear(damage);
@@ -1122,6 +1192,24 @@ void wlr_surface_get_effective_damage(struct wlr_surface *surface,
 		surface->current.buffer_height);
 	wlr_region_scale(damage, damage, 1.0 / (float)surface->current.scale);
 
+	if (surface->current.viewport.has_src) {
+		struct wlr_box src_box = {
+			.x = floor(surface->current.viewport.src.x),
+			.y = floor(surface->current.viewport.src.y),
+			.width = ceil(surface->current.viewport.src.width),
+			.height = ceil(surface->current.viewport.src.height),
+		};
+		crop_region(damage, damage, &src_box);
+	}
+	if (surface->current.viewport.has_dst) {
+		int src_width, src_height;
+		surface_state_viewport_src_size(&surface->current,
+			&src_width, &src_height);
+		float scale_x = (float)surface->current.viewport.dst_width / src_width;
+		float scale_y = (float)surface->current.viewport.dst_height / src_height;
+		wlr_region_scale_xy(damage, damage, scale_x, scale_y);
+	}
+
 	// On resize, damage the previous bounds of the surface. The current bounds
 	// have already been damaged in surface_update_damage.
 	if (surface->previous.width > surface->current.width ||
@@ -1143,3 +1231,26 @@ void wlr_surface_get_effective_damage(struct wlr_surface *surface,
 			surface->previous.width, surface->previous.height);
 	}
 }
+
+void wlr_surface_get_buffer_source_box(struct wlr_surface *surface,
+		struct wlr_fbox *box) {
+	box->x = box->y = 0;
+	box->width = surface->current.buffer_width;
+	box->height = surface->current.buffer_height;
+
+	if (surface->current.viewport.has_src) {
+		box->x = surface->current.viewport.src.x * surface->current.scale;
+		box->y = surface->current.viewport.src.y * surface->current.scale;
+		box->width = surface->current.viewport.src.width * surface->current.scale;
+		box->height = surface->current.viewport.src.height * surface->current.scale;
+		if ((surface->current.transform & WL_OUTPUT_TRANSFORM_90) != 0) {
+			double tmp = box->x;
+			box->x = box->y;
+			box->y = tmp;
+
+			tmp = box->width;
+			box->width = box->height;
+			box->height = tmp;
+		}
+	}
+}
