@@ -25,6 +25,10 @@ static void xdg_pointer_grab_enter(struct wlr_seat_pointer_grab *grab,
 	}
 }
 
+static void xdg_pointer_grab_clear_focus(struct wlr_seat_pointer_grab *grab) {
+	wlr_seat_pointer_clear_focus(grab->seat);
+}
+
 static void xdg_pointer_grab_motion(struct wlr_seat_pointer_grab *grab,
 		uint32_t time, double sx, double sy) {
 	wlr_seat_pointer_send_motion(grab->seat, time, sx, sy);
@@ -59,6 +63,7 @@ static void xdg_pointer_grab_cancel(struct wlr_seat_pointer_grab *grab) {
 
 static const struct wlr_pointer_grab_interface xdg_pointer_grab_impl = {
 	.enter = xdg_pointer_grab_enter,
+	.clear_focus = xdg_pointer_grab_clear_focus,
 	.motion = xdg_pointer_grab_motion,
 	.button = xdg_pointer_grab_button,
 	.cancel = xdg_pointer_grab_cancel,
@@ -72,6 +77,10 @@ static void xdg_keyboard_grab_enter(struct wlr_seat_keyboard_grab *grab,
 	// keyboard focus should remain on the popup
 }
 
+static void xdg_keyboard_grab_clear_focus(struct wlr_seat_keyboard_grab *grab) {
+	// keyboard focus should remain on the popup
+}
+
 static void xdg_keyboard_grab_key(struct wlr_seat_keyboard_grab *grab, uint32_t time,
 		uint32_t key, uint32_t state) {
 	wlr_seat_keyboard_send_key(grab->seat, time, key, state);
@@ -88,6 +97,7 @@ static void xdg_keyboard_grab_cancel(struct wlr_seat_keyboard_grab *grab) {
 
 static const struct wlr_keyboard_grab_interface xdg_keyboard_grab_impl = {
 	.enter = xdg_keyboard_grab_enter,
+	.clear_focus = xdg_keyboard_grab_clear_focus,
 	.key = xdg_keyboard_grab_key,
 	.modifiers = xdg_keyboard_grab_modifiers,
 	.cancel = xdg_keyboard_grab_cancel,
@@ -278,11 +288,6 @@ void create_xdg_popup(struct wlr_xdg_surface *xdg_surface,
 		return;
 	}
 
-	if (!wlr_surface_set_role(xdg_surface->surface, &xdg_popup_surface_role,
-			xdg_surface, xdg_surface->resource, XDG_WM_BASE_ERROR_ROLE)) {
-		return;
-	}
-
 	if (xdg_surface->role != WLR_XDG_SURFACE_ROLE_NONE) {
 		wl_resource_post_error(xdg_surface->resource,
 			XDG_SURFACE_ERROR_ALREADY_CONSTRUCTED,
@@ -290,6 +295,11 @@ void create_xdg_popup(struct wlr_xdg_surface *xdg_surface,
 		return;
 	}
 
+	if (!wlr_surface_set_role(xdg_surface->surface, &xdg_popup_surface_role,
+			xdg_surface, xdg_surface->resource, XDG_WM_BASE_ERROR_ROLE)) {
+		return;
+	}
+
 	assert(xdg_surface->popup == NULL);
 	xdg_surface->popup = calloc(1, sizeof(struct wlr_xdg_popup));
 	if (!xdg_surface->popup) {
@@ -454,10 +464,10 @@ static bool xdg_popup_unconstrain_flip(struct wlr_xdg_popup *popup,
 	}
 
 	// revert the positioner back if it didn't fix it and go to the next part
-	if (flip_x) {
+	if (offset_x && flip_x) {
 		wlr_positioner_invert_x(&popup->positioner);
 	}
-	if (flip_y) {
+	if (offset_y && flip_y) {
 		wlr_positioner_invert_y(&popup->positioner);
 	}
 
