@@ -35,6 +35,10 @@ static void xdg_pointer_grab_enter(struct wlr_seat_pointer_grab *grab,
 	}
 }
 
+static void xdg_pointer_grab_clear_focus(struct wlr_seat_pointer_grab *grab) {
+	wlr_seat_pointer_clear_focus(grab->seat);
+}
+
 static void xdg_pointer_grab_motion(struct wlr_seat_pointer_grab *grab,
 		uint32_t time, double sx, double sy) {
 	wlr_seat_pointer_send_motion(grab->seat, time, sx, sy);
@@ -69,6 +73,7 @@ static void xdg_pointer_grab_cancel(struct wlr_seat_pointer_grab *grab) {
 
 static const struct wlr_pointer_grab_interface xdg_pointer_grab_impl = {
 	.enter = xdg_pointer_grab_enter,
+	.clear_focus = xdg_pointer_grab_clear_focus,
 	.motion = xdg_pointer_grab_motion,
 	.button = xdg_pointer_grab_button,
 	.cancel = xdg_pointer_grab_cancel,
@@ -82,6 +87,10 @@ static void xdg_keyboard_grab_enter(struct wlr_seat_keyboard_grab *grab,
 	// keyboard focus should remain on the popup
 }
 
+static void xdg_keyboard_grab_clear_focus(struct wlr_seat_keyboard_grab *grab) {
+	// keyboard focus should remain on the popup
+}
+
 static void xdg_keyboard_grab_key(struct wlr_seat_keyboard_grab *grab,
 		uint32_t time, uint32_t key, uint32_t state) {
 	wlr_seat_keyboard_send_key(grab->seat, time, key, state);
@@ -98,6 +107,7 @@ static void xdg_keyboard_grab_cancel(struct wlr_seat_keyboard_grab *grab) {
 
 static const struct wlr_keyboard_grab_interface xdg_keyboard_grab_impl = {
 	.enter = xdg_keyboard_grab_enter,
+	.clear_focus = xdg_keyboard_grab_clear_focus,
 	.key = xdg_keyboard_grab_key,
 	.modifiers = xdg_keyboard_grab_modifiers,
 	.cancel = xdg_keyboard_grab_cancel,
@@ -474,10 +484,10 @@ static bool xdg_popup_v6_unconstrain_flip(struct wlr_xdg_popup_v6 *popup,
 	}
 
 	// revert the positioner back if it didn't fix it and go to the next part
-	if (flip_x) {
+	if (offset_x && flip_x) {
 		wlr_positioner_v6_invert_x(&popup->positioner);
 	}
-	if (flip_y) {
+	if (offset_y && flip_y) {
 		wlr_positioner_v6_invert_y(&popup->positioner);
 	}
 
