@@ -8,36 +8,55 @@
 #include <unistd.h>
 #include <wayland-server-core.h>
 #include <wlr/util/log.h>
+#include "util/time.h"
 
 static bool colored = true;
 static enum wlr_log_importance log_importance = WLR_ERROR;
+static struct timespec start_time = {-1};
 
 static const char *verbosity_colors[] = {
 	[WLR_SILENT] = "",
-	[WLR_ERROR ] = "\x1B[1;31m",
-	[WLR_INFO  ] = "\x1B[1;34m",
-	[WLR_DEBUG ] = "\x1B[1;30m",
+	[WLR_ERROR] = "\x1B[1;31m",
+	[WLR_INFO] = "\x1B[1;34m",
+	[WLR_DEBUG] = "\x1B[1;90m",
 };
 
+static const char *verbosity_headers[] = {
+	[WLR_SILENT] = "",
+	[WLR_ERROR] = "[ERROR]",
+	[WLR_INFO] = "[INFO]",
+	[WLR_DEBUG] = "[DEBUG]",
+};
+
+static void init_start_time(void) {
+	if (start_time.tv_sec >= 0) {
+		return;
+	}
+	clock_gettime(CLOCK_MONOTONIC, &start_time);
+}
+
 static void log_stderr(enum wlr_log_importance verbosity, const char *fmt,
 		va_list args) {
+	init_start_time();
+
 	if (verbosity > log_importance) {
 		return;
 	}
-	// prefix the time to the log message
-	struct tm result;
-	time_t t = time(NULL);
-	struct tm *tm_info = localtime_r(&t, &result);
-	char buffer[26];
 
-	// generate time prefix
-	strftime(buffer, sizeof(buffer), "%F %T - ", tm_info);
-	fprintf(stderr, "%s", buffer);
+	struct timespec ts = {0};
+	clock_gettime(CLOCK_MONOTONIC, &ts);
+	timespec_sub(&ts, &ts, &start_time);
+
+	fprintf(stderr, "%02d:%02d:%02d.%03ld ", (int)(ts.tv_sec / 60 / 60),
+		(int)(ts.tv_sec / 60 % 60), (int)(ts.tv_sec % 60),
+		ts.tv_nsec / 1000000);
 
 	unsigned c = (verbosity < WLR_LOG_IMPORTANCE_LAST) ? verbosity : WLR_LOG_IMPORTANCE_LAST - 1;
 
 	if (colored && isatty(STDERR_FILENO)) {
 		fprintf(stderr, "%s", verbosity_colors[c]);
+	} else {
+		fprintf(stderr, "%s ", verbosity_headers[c]);
 	}
 
 	vfprintf(stderr, fmt, args);
@@ -60,6 +79,8 @@ static void log_wl(const char *fmt, va_list args) {
 }
 
 void wlr_log_init(enum wlr_log_importance verbosity, wlr_log_func_t callback) {
+	init_start_time();
+
 	if (verbosity < WLR_LOG_IMPORTANCE_LAST) {
 		log_importance = verbosity;
 	}
